/**
 * Buffer like StringBuffer for byte array.
 *
 * @author Heiner K&uuml;cker
 */
public class ByteStrBuff
{
    private byte[] bArr;

    /** Pointer to the next free position */
    private int size = 0;

    /**
     * Constructor with capacity.
     * Capacitiy must be a value greater zero.
     * 
     * @param pCapacity
     */
    public ByteStrBuff(final int pCapacity)
    {
        bArr = new byte[ pCapacity ];
    }

    /**
     * Constructor with specified byte array.
     * Attention, no side effects safe.
     * 
     * @param pArr
     */
    public ByteStrBuff(final byte[] pArr)
    {
        bArr = pArr;
        size = pArr.length;
    }

    /**
     * Constructor with byte array and capacitiy.
     * 
     * @param pArr
     * @param pCapacity
     */
    public ByteStrBuff(//-
        final byte[] pArr, //--
        final int pCapacity)
    {
        bArr = new byte[ Math.max( pArr.length , pCapacity ) ];
        System.arraycopy(pArr, 0, bArr, 0, pArr.length);
        size = pArr.length;
    }

    /**
     * @return
     */
    public final int length()
    {
        return size;
    }

    /**
     * Set the pointer in this
     * Str8Buff to initial position
     * for reusing this.
     */
    public void reset()
    {
        size = 0;
    }

    /**
     * Get the capacity.
     * 
     * @return
     */
    public final int capacity()
    {
        return bArr.length;
    }

    /**
     * Set the capacity.
     * 
     * @param pCapacity new capacitiy
     * @return new capacitiy
     */
    public final int capacity(final int pCapacity)
    {
        if ( pCapacity > size)
        {
            byte[] tArr = bArr;
            bArr = new byte[ pCapacity ];
            System.arraycopy(tArr, 0, bArr, 0, tArr.length);

        }
        return bArr.length;
    }

    /**
     * Get the byte at the specified position.
     * 
     * @param index
     * @return
     */
    public final byte byteAt(final int index)
    {
        if ( index >= size )
          {
          throw new ArrayIndexOutOfBoundsException( "index: " + index + " not useable for size: " + size );
          }
        return bArr[index];
    }

    /**
     * Get a clone of the saved bytes.
     * 
     * @return
     */
    public final byte[] bytes()
    {
        byte[] retArr = new byte[ size ];
        System.arraycopy(bArr, 0, retArr, 0, size);
        return retArr;
    }

    /**
     * Append a byte at the next fre position.
     * Resize the capacity if neccessary.
     * 
     * @param pByte byte to append
     */
    public final void append(final byte pByte)
    {
        if (size >= bArr.length)
        {
            capacity( (size + 1) * 2);
        }
        bArr[ size ] = pByte;
        size++;
    }

    /**
     * Append a byte array at the next fre position.
     * Resize the capacity if neccessary.
     * 
     * @param pBytArr byte array to append
     */
    public final void append(final byte[] pBytArr)
    {
        if (size + pBytArr.length >= bArr.length)
        {
            capacity( (size + pBytArr.length + 1) * 2);
        }
        System.arraycopy(pBytArr, 0, bArr, size, pBytArr.length);
        size += pBytArr.length;
    }

    /**
     *
     */
    public final String toString()
    {
       return new String( bytes() );
    }

    /**
     *
     * @param pArr
     * @param pOldArr
     * @param pNewArr
     * @return
     */
    public static final byte[] replace( //--
        final byte[] pArr, //--
        final byte[] pOldArr, //--
        final byte[] pNewArr)
    {
        final int arrLen = pArr.length;
        final ByteStrBuff buff = new ByteStrBuff( (int) ( arrLen * 1.5 ) );
        for (int i = 0; i < arrLen; i++)
        {
            if ( ByteStr.startsWith(pArr, pOldArr, i) )
            {
                buff.append( pNewArr );
                i += pOldArr.length - 1;
            }
            else
            {
                buff.append( pArr[i] );
            }
        }
        return buff.bytes();
    }

    /**
     * @param S430_pByteArr
     * @return
     */
    public final boolean startsWith(final byte[] pByteArr)
    {
        if ( pByteArr.length > size )
        {
            return false;
        }
        return ByteStr.startsWith(bArr, pByteArr);
    }

    /**
     * @param pSize
     */
    public final byte[] leftLenient(final int pSize)
    {
        return ByteStr.leftLenient(bArr, Math.min( pSize , size ) );
    }

    /**
     * @param bs
     * @return
     */
    public final int indexOf(final byte[] pByteArr)
    {
        return ByteStr.indexOf(bArr, pByteArr, 0, size);
    }

    /**
     * @param
     * @param pFromPos
     * @return
     */
    public final int indexOf( //--
    final byte[] pByteArr, //--
    final int pFromPos)
    {
        return ByteStr.indexOf(bArr, pByteArr, pFromPos, size);
    }

    /**
     * @param startPos
     * @param endPos
     * @return
     */
    public final byte[] sub(//--
        final int pFromPos, //--
        final int pToPos)
    {
        if ( pToPos > size )
        {
            throw new ArrayIndexOutOfBoundsException(pToPos);
        }
        return ByteStr.sub(bArr, pFromPos, pToPos );
    }

    /**
     * @param oldPos
     * @param LOG_SIZE
     */
    public final byte[] subLenient(//--
        final int pFromPos, //--
        final int pToPos)
    {
        if ( pFromPos >= this.size )
        {
            return new byte[ 0 ];
        }
        return sub( pFromPos , Math.min( size , pToPos ) );
    }
    /**
     * @param
     * @return
     */
    public final boolean endsWith(final byte[] pArr)
    {
        return ByteStr.endsWith(bArr, pArr, size);
    }

    /**
     * Equals.
     */
    public boolean equals(Object o)
    {
      if (o.getClass() != this.getClass())
      {
        return false;
      }
      final ByteStrBuff otherBuf = (ByteStrBuff) o;
      if (this.size != otherBuf.size)
      {
        return false;
      }
      
      // this is an error,if the unused space is different
      //return Arrays.equals(otherBuf.bytes(), this.bytes());

      final byte[] thisByteArr = this.bArr;
      final byte[] otherByteArr = otherBuf.bArr;
      for (int i = 0; i < this.size; i++)
      {
        if ( thisByteArr[i] != otherByteArr[i] )
        {
          return false;
        }
      }
      return true;
    }

    /**
     * Test.
     * @param args
     */
    public static void main(String[] args)
    {
        byte[] bArr = "..abc\r\n..a123\r\n..\r\n.".getBytes();
        bArr = replace(bArr, "\r\n..".getBytes(), "\r\n.".getBytes() );
        System.out.println( new String( bArr ) );
    }

}
