import java.io.IOException;
import java.util.Arrays;

/**
 * XML really pull parser.
 * 
 * Ein XML-Parser, der diesen Namen auch verdient.
 * 
 * Ein ziemlich schneller XML Parser. Weniger Flexibilitt, aber hhere Performance.
 * 
 * @author Heiner Kcker
 */
public final class XrppParser
{
  /** globaler Schalter, ob Kommentare unterstzt werden */
  private static final boolean IS_COMMENT_SUPPORTED = true;

  private final MultiMarkableBufferedInputStream inStream;

  /**
   * Constructor with InputStream to parse.
   * 
   * @param pInStream
   */
  public XrppParser(// --
      final MultiMarkableBufferedInputStream pInStream)
  {
    this.inStream = pInStream;
  }

  /**
   * berspringen weisse Zeichen.
   * 
   * TODO berspringen Kommentare
   * 
   * @param pBuffInStream
   * @throws IOException
   * @throws XrppParseException
   */
  private final void skipWhiteCharsAndComments()
      throws IOException ,
      XrppParseException
  {
    while (true)
    {
      MultiMarkableBufferedInputStream.Mark mark = inStream.getMark();

      byte nextByte = (byte) inStream.read();

      // System.out.println( "nextByte: " + (char) nextByte );

      if (nextByte == -1)
      {
        // System.out.println( "end of file reached" );
        return;
      }

      if (IS_COMMENT_SUPPORTED
          && nextByte == (byte) '<')
      {
        // eventually an comment
        // if ( (byte) inStream.read() == (byte) '!'
        // && (byte) inStream.read() == (byte) '-'
        // && (byte) inStream.read() == (byte) '-' )
        if (isByteSequence( "!--".getBytes() ))
        {
          // is an comment
          System.out.println( "skip comment "
              + mark.getPosFromStart() );
          while (true)
          {
            MultiMarkableBufferedInputStream.Mark commentMark = inStream.getMark();
            if (isByteSequence( "-->".getBytes() ))
            {
              System.out.println( "end comment "
                  + inStream.getPosFromStart() );
              break;
            }
            inStream.resetMark( commentMark );
            if (inStream.read() < 0)
            {
              throw new XrppParseException( "unclosed comment" );
            }
          }

        }
        else
        {
          inStream.resetMark( mark );
          return;
        }
      }
      else if ( !Character.isWhitespace( (char) nextByte ))
      {
        // inStream.reset();
        inStream.resetMark( mark );
        return;
      }
    }
  }

  /**
   * @param pBuffInStream
   * @return
   * @throws IOException
   */
  private byte skipWhiteCharsNoMark()
      throws IOException
  {
    while (true)
    {
      // TODO Beschleunigen durch direkten byte vergleich statt char
      byte nextByte = (byte) inStream.read();
      if ( !Character.isWhitespace( (char) nextByte ))
      {
        return nextByte;
      }
    }
  }

  /**
   * Lese Zeichen und wirf Fehler, wenn Zeischen nicht da.
   * 
   * @param pBuffInStream
   * @param pByte
   *          expected byte
   * @throws IOException
   * @throws XrppParseException
   */
  private void forceByte(
      final byte pByte )
      throws IOException ,
      XrppParseException
  {
    final byte nextByte = (byte) inStream.read();
    if (nextByte != pByte)
    {
      throw new XrppParseException( ( (char) pByte )
          + " expected instead '" + (char) nextByte + "' at pos " + inStream.getPosFromStart() );
    }
  }

  /**
   * @param pBuffInStream
   * @param pByteArr
   * @throws IOException
   * @throws XrppParseException
   */
  private void forceByteSequence(
      final byte[] pByteArr )
      throws IOException ,
      XrppParseException
  {
    // for (int i = 0; i < pByteArr.length; i++)
    // {
    // byte by = pByteArr[i];
    // forceByte(inStream, by);
    // }
    final byte[] tByteArr = new byte[pByteArr.length];
    final int size = inStream.read( tByteArr );
    if (size != pByteArr.length
        || !Arrays.equals( pByteArr , tByteArr ))
    {
      throw new XrppParseException( "sequence: "
          + new String( pByteArr ) + " expected, found " + new String( tByteArr ) );
    }
  }

  /**
   * Verlange Element Start, Element kann leer oder nicht leer sein.
   * 
   * Schneller mit element name als byte array statt string
   * 
   * @param inStream
   * @param pByteArr
   *          element name as byte array
   * @throws IOException
   * @throws XrppParseException
   */
  public void forceElementStart(
      final byte[] pByteArr )
      throws IOException ,
      XrppParseException
  {
    System.out.println( "forceElementStart: "
        + new String( pByteArr ) + " " + inStream.getPosFromStart() );

    skipWhiteCharsAndComments();

    forceByte( (byte) '<' );

    skipWhiteCharsAndComments();

    forceByteSequence( pByteArr );
  }

  /**
   * Gehe zum Element Body eines nicht leeren Elementes.
   * 
   * @param inStream
   * @throws IOException
   * @throws XrppParseException
   */
  public void forceElementBody()
      throws IOException ,
      XrppParseException
  {
    try
    {
      skipWhiteCharsAndComments();

      forceByte( (byte) '>' );
    }
    catch (XrppParseException e)
    {
      System.err.println( e.getMessage() );
      e.printStackTrace();

      throw new XrppParseException( "ElementBody expected" );
    }
  }

  /**
   * Verlange Element Stop, Element darf nicht leer sein.
   * 
   * Schneller mit element name als byte array statt string
   * 
   * @param inStream
   * @param pByteArr
   *          element name as byte array
   * @throws IOException
   * @throws IOException
   * @throws XrppParseException
   */
  public void forceElementStopNoEmpty(
      final byte[] pByteArr )
      throws IOException ,
      XrppParseException
  {
System.out.println( "forceElementStopNoEmpty" );    
    try
    {
      skipWhiteCharsAndComments();

      forceByte( (byte) '<' );

      skipWhiteCharsAndComments();

      forceByte( (byte) '/' );

      skipWhiteCharsAndComments();

      forceByteSequence( pByteArr );

      skipWhiteCharsAndComments();

      forceByte( (byte) '>' );
    }
    catch (XrppParseException e)
    {
      System.err.println( e.getMessage() );
      e.printStackTrace();

      throw new XrppParseException( "ElementStopNoEmpty expected: "
          + new String( pByteArr ) );
    }
  }

  /**
   * Verlange Element Stop, Element darf leer oder nicht leer sein.
   * 
   * Schneller mit element name als byte array statt string
   * 
   * @param inStream
   * @param pElementNameByteArr
   *          element name as byte array
   * @throws IOException
   * @throws IOException
   * @throws XrppParseException
   */
  public void forceElementStopEmptyOrNoEmpty(
      final byte[] pElementNameByteArr )
      throws IOException ,
      XrppParseException
  {
    try
    {
      skipWhiteCharsAndComments();

      // Test fr Tag-Ende mit />
      MultiMarkableBufferedInputStream.Mark mark = inStream.getMark();

      if ((byte) inStream.read() == (byte) '/')
      {
        skipWhiteCharsAndComments();
        forceByte( (byte) '>' );
        return;
      }
      else
      {
        inStream.resetMark( mark );
      }

      // Test fr Tag-Ende mit ></name>
      forceByte( (byte) '>' );

      skipWhiteCharsAndComments();

      forceByte( (byte) '<' );

      skipWhiteCharsAndComments();

      forceByte( (byte) '/' );

      skipWhiteCharsAndComments();

      forceByteSequence( pElementNameByteArr );

      skipWhiteCharsAndComments();

      forceByte( (byte) '>' );
    }
    catch (XrppParseException e)
    {
      System.err.println( e.getMessage() );
      e.printStackTrace();

      throw new XrppParseException( "ElementStopEmptyOrNoEmpty expected: "
          + new String( pElementNameByteArr ) );
    }
  }

  /**
   * Verlange Element Stop, Element darf nicht leer sein.
   * 
   * Schneller mit element name als byte array statt string
   * 
   * @param inStream
   * @param pAttributeNameByteArr
   *          attribute name as byte array
   * @throws IOException
   * @throws IOException
   * @throws XrppParseException
   */
  public byte[] getAttribute(
      final byte[] pAttributeNameByteArr )
      throws IOException ,
      XrppParseException
  {
    skipWhiteCharsAndComments();

    forceByteSequence( pAttributeNameByteArr );

    skipWhiteCharsAndComments();

    forceByte( (byte) '=' );

    skipWhiteCharsAndComments();

    forceByte( (byte) '"' );

    final ByteStrBuff buff = new ByteStrBuff( 64 );
    byte by;
    while (( by = (byte) inStream.read() ) != (byte) '"')
    {
      buff.append( by );
    }

    System.out.println( "getAttribute "
        + new String( pAttributeNameByteArr ) + ": " + new String( buff.bytes() ) );

    return buff.bytes();
  }

  /**
   * @param inStream
   * @param pBytes
   * @return
   */
  public boolean isElementStart(
      final byte[] pByteArr )
      throws IOException ,
      XrppParseException
  {
    System.out.print( "isElementStart: "
        + new String( pByteArr ) + " " + inStream.getPosFromStart() + ": " );
    MultiMarkableBufferedInputStream.Mark mark = inStream.getMark();

    skipWhiteCharsAndComments();
    if ((byte) inStream.read() != (byte) '<')
    {
      inStream.resetMark( mark );
      System.out.println( false );
      return false;
    }

    if (skipWhiteCharsNoMark() != pByteArr[0])
    {
      // inStream.reset();
      inStream.resetMark( mark );
      System.out.println( false );
      return false;
    }

    for (int i = 1; i < pByteArr.length; i++ )
    {
      byte by = pByteArr[i];
      if (by != (byte) inStream.read())
      {
        // inStream.reset();
        inStream.resetMark( mark );
        System.out.println( false );
        return false;
      }
    }

    byte by = (byte) inStream.read();
    if (( !Character.isWhitespace( (char) by ) )
        && by != (byte) '>' 
            && by != (byte) '/')
    {
      // inStream.reset();
      inStream.resetMark( mark );
      System.out.println( false );
      return false;
    }
    // inStream.reset();
    inStream.resetMark( mark );

    System.out.println( true );
    return true;
  }

  /**
   * Prfe, ob an der aktuellen Stelle im InputStream das bergebene byte array steht.
   * 
   * @param inStream
   * @param pByteArr
   * @return
   * @throws IOException
   */
  private final boolean isByteSequence(
      final byte[] pByteArr )
      throws IOException
  {
    for (int i = 0; i < pByteArr.length; i++ )
    {
      final byte forceByte = pByteArr[i];
      final boolean isByte = isByte( forceByte );
      if (! isByte)
      {
        return false;
      }
    }
    return true;
  }

  /**
   * Prfe, ob an der aktuellen Stelle im InputStream das bergebene byte steht.
   * 
   * @param pForceByte
   * @return
   * @throws IOException
   */
  private boolean isByte(
      final byte pForceByte )
      throws IOException
  {
    final byte readByte = (byte) inStream.read();
    final boolean isByte = readByte == pForceByte;
    //System.out.println( "isByte: '" + (char) pForceByte + "' " + isByte );
    return isByte;
  }

  /**
   * berspringe XMl-Deklaration, wenn vorhanden.
   * 
   * @param inStream
   * @throws IOException
   * @throws XrppParseException
   */
  public void skipXmlDeclaration()
      throws IOException ,
      XrppParseException
  {
    System.out.println( "skipXmlDeclaration: "
        + inStream.getPosFromStart() );

    skipWhiteCharsAndComments();

    MultiMarkableBufferedInputStream.Mark mark = inStream.getMark();

    if (isByteSequence( "<?".getBytes() ))
    {
      // bis zum '?' laufen
      int readInt = -1;
      while ((byte) ( readInt = inStream.read() ) != (byte) '?')
      {
        if (readInt == -1)
        {
          throw new XrppParseException( "unexpected end of document" );
        }
      }
      // '>' lesen
      forceByte( (byte) '>' );
    }
    else
    {
      // kein '<?' gefunden
      inStream.resetMark( mark );
    }

  }

  /**
   * berspringe Document-Type-Deklaration, wenn vorhanden.
   * 
   * @param inStream
   * @throws IOException
   * @throws XrppParseException
   */
  public void skipDoctypeDeclaration()
      throws IOException ,
      XrppParseException
  {
    System.out.println( "skipDoctypeDeclaration: "
        + inStream.getPosFromStart() );

    skipWhiteCharsAndComments();

    MultiMarkableBufferedInputStream.Mark mark = inStream.getMark();

    if (isByteSequence( "<!".getBytes() ))
    {
      // bis zum '>' laufen
      int readInt = -1;
      while ((byte) ( readInt = inStream.read() ) != (byte) '>')
      {
        if (readInt == -1)
        {
          throw new XrppParseException( "unexpected end of document" );
        }
      }
    }
    else
    {
      // kein '<?' gefunden
      inStream.resetMark( mark );
    }
  }

  /**
   * Lies fr ein eventuell vorhandenes Element Element-Start, Element-Body und Eelement-Ende und gib den Element-Body
   * zurck. Das Element darf keine Sub-Elemente enthalten.
   * 
   * @param inStream
   * @param pBytes
   * @throws XrppParseException
   * @throws IOException
   */
  public byte[] getOptionalElementBody(
      final byte[] pElementNameByteArr )
      throws IOException ,
      XrppParseException
  {
    if (isElementStart( pElementNameByteArr ))
    {
      return getElementBodyWithStartAndEndTag( pElementNameByteArr );
    }
    return null;
  }

  /**
   * Lies Element-Start, Element-Body und Eelement-Ende und gib den Element-Body zurck. Das Element darf keine
   * Sub-Elemente enthalten.
   * 
   * @param inStream
   * @param pBytes
   * @throws XrppParseException
   * @throws IOException
   */
  public byte[] getElementBodyWithStartAndEndTag(
      final byte[] pElementNameByteArr )
      throws IOException ,
      XrppParseException
  {
    forceElementStart( pElementNameByteArr );
    ByteStrBuff byteStrBuff = null;
    try
    {
      skipWhiteCharsAndComments();

      // Test fr Tag-Ende mit />
      MultiMarkableBufferedInputStream.Mark mark = inStream.getMark();

      if ((byte) inStream.read() == (byte) '/')
      {
        // Element ist empty
        skipWhiteCharsAndComments();
        forceByte( (byte) '>' );
        return null;
      }
      else
      {
        inStream.resetMark( mark );
      }

      // Test fr Tag-Ende mit ></name>
      forceByte( (byte) '>' );

      // skipWhiteCharsAndComments( inStream );
      // forceByte( inStream , (byte) '<' );

      // bis zum '<' laufen und element body lesen
      int readInt = -1;
      byteStrBuff = new ByteStrBuff( 1024 );
      while ((byte) ( readInt = inStream.read() ) != (byte) '<')
      {
        if (readInt == -1)
        {
          throw new XrppParseException( "unexpected end of document" );
        }
        byteStrBuff.append( (byte) readInt );
      }

      System.out.println( "getElementBody "
          + new String( pElementNameByteArr ) + ": " + new String( byteStrBuff.bytes() ) );

      skipWhiteCharsAndComments();

      forceByte( (byte) '/' );

      skipWhiteCharsAndComments();

      forceByteSequence( pElementNameByteArr );

      skipWhiteCharsAndComments();

      forceByte( (byte) '>' );
    }
    catch (XrppParseException e)
    {
      System.err.println( e.getMessage() );
      e.printStackTrace();

      throw new XrppParseException( "ElementStopEmptyOrNoEmpty expected: "
          + new String( pElementNameByteArr ) );
    }
    return byteStrBuff.bytes();
  }

  /**
   * Lies Element-Body ohne beginnendes Tags. Vorrausssetzung ist der vorherige Aufruf von forceElementStart() und das
   * Lesen aller Attriibute.
   */
  public byte[] getElementBodyWithoutStart(
      final byte[] pElementNameByteArr )
      throws IOException ,
      XrppParseException
  {
    // hier nicht: forceElementStart( pElementNameByteArr );
    ByteStrBuff byteStrBuff = null;
    try
    {
      skipWhiteCharsAndComments();

      // Test fr Tag-Ende mit />
      MultiMarkableBufferedInputStream.Mark mark = inStream.getMark();

      if ((byte) inStream.read() == (byte) '/')
      {
        // Element ist empty
        skipWhiteCharsAndComments();
        forceByte( (byte) '>' );
        return null;
      }
      else
      {
        inStream.resetMark( mark );
      }

      // Test fr Tag-Ende mit ></name>
      forceByte( (byte) '>' );

      // skipWhiteCharsAndComments( inStream );
      // forceByte( inStream , (byte) '<' );

      // bis zum '<' laufen und element body lesen
      int readInt = -1;
      byteStrBuff = new ByteStrBuff( 1024 );
      while ((byte) ( readInt = inStream.read() ) != (byte) '<')
      {
        if (readInt == -1)
        {
          throw new XrppParseException( "unexpected end of document" );
        }
        byteStrBuff.append( (byte) readInt );
      }

      System.out.println( "getElementBodyWithoutStart "
          + new String( pElementNameByteArr ) + ": " + new String( byteStrBuff.bytes() ) );

      skipWhiteCharsAndComments();

      forceByte( (byte) '/' );

      skipWhiteCharsAndComments();

      forceByteSequence( pElementNameByteArr );

      skipWhiteCharsAndComments();

      forceByte( (byte) '>' );
    }
    catch (XrppParseException e)
    {
      System.err.println( e.getMessage() );
      e.printStackTrace();

      throw new XrppParseException( "ElementStopEmptyOrNoEmpty expected: "
          + new String( pElementNameByteArr ) );
    }
    return byteStrBuff.bytes();
  }

  /**
   * Prfe, ob alle Daten aus dem XML-InputStream gelesen wurden.
   * 
   * @throws XrppParseException
   * @throws IOException
   */
  public void checkForEndOfDocument()
      throws IOException ,
      XrppParseException
  {
    skipWhiteCharsAndComments();

    if (inStream.read() != -1)
    {
      throw new XrppParseException( "end of document expected: "
          + inStream.getPosFromStart() );
    }
  }

  /**
   * Prfe, ob das aktuelle Element leer ist (ohne Verschieben der Markierung im InputStream).
   * 
   * @return
   * @throws XrppParseException
   * @throws IOException
   */
  public boolean isElementEmpty()
      throws IOException ,
      XrppParseException
  {

    System.out.println( "isElementEmpty " + inStream.getPosFromStart() );
    
    // Test fr Tag-Ende mit />
    MultiMarkableBufferedInputStream.Mark mark = inStream.getMark();

    skipWhiteCharsAndComments();

    if (isByte( (byte) '/' ))
    {
//System.out.println( "/ ist da");      
      // Element ist empty
      skipWhiteCharsAndComments();
      if (isByte( (byte) '>' ))
      {
//System.out.println( "> ist da");      
        inStream.resetMark( mark );
        System.out.println( "isElementEmpty " + true );
        return true;
      }
    }
    inStream.resetMark( mark );
    
    System.out.println( "isElementEmpty " + false );
    return false;
  }

  /**
   * Verbrauche das leere Element-Ende '/>'.
   * @throws XrppParseException 
   * @throws IOException 
   */
  public void forceElementStopEmpty() 
  throws IOException, XrppParseException
  {
    System.out.println( "forceElementStopEmpty" );
    // Verlange Tag-Ende mit />
    skipWhiteCharsAndComments();

    if (isByte( (byte) '/') )
    {
      // Element ist empty
      skipWhiteCharsAndComments();
      if (isByte( (byte) '>' ))
      {
        return;
      }
    }
    throw new XrppParseException( "end of empty element expected " + inStream.getPosFromStart() );
  }

}
